/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems, Cider Press
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Vpf/vpfdb.h>
#include <Vpf/vpflib.h>

#include <iostream.h>
#include <string.h>
#include <sys/stat.h>

/*****************************************************************************/

VpfDatabase::VpfDatabase (const char* path, boolean initialize) 
: MapDatabase(nil, path) {
    const char *file = strrchr(path, '/');
    _name = strdup(strdup (file ? file : path));
    _dbprod = NULL;
    _dbdesc = NULL;
    if (initialize)
        init();
}

VpfDatabase::~VpfDatabase () {
    delete _dbprod;
}

void VpfDatabase::init() {
    producer();
    init_libraries();
}

VpfLibrary* VpfDatabase::library(unsigned libnum) const {
    return (VpfLibrary*)MapDatabase::library(libnum);
}

VpfLibrary* VpfDatabase::library(const char* name) const {
    return (VpfLibrary*)MapDatabase::library(name);
}

const char* VpfDatabase::producer() const {
    if (!_dbprod) {
	vpf_table_type table;
	row_type row;
	char dhtpath[255];
	strcpy(dhtpath, path());
	strcat(dhtpath, "/dht.");
	table = vpf_open_table(dhtpath, disk, "rb", NULL);
	if (!table.fp) {
	    return nil;
	}
	else {
	    long int prodfieldnum = table_pos("ORIGINATOR", table);
	    if (prodfieldnum < 0) {
		cerr << "VpfDatabase::producer: ";
		cerr << "Invalid DHT (" << path()
		     << ") - missing ORIGINATOR field\n";
		vpf_close_table(&table);
	    }
	    else {
		long int n;
		row = read_next_row(table);
		((VpfDatabase*)this)->_dbprod = (char *) get_table_element(prodfieldnum,row,table,NULL,&n);
		free_row(row,table);
		vpf_close_table(&table);
	    }      
	}
    }
    return _dbprod;
}

const char* VpfDatabase::description() const {
    if (!_dbdesc) {
	vpf_table_type table;
	row_type row;
	char dhtpath[255];
	strcpy(dhtpath, path());
	strcat(dhtpath, "/dht.");
	table = vpf_open_table(dhtpath, disk, "rb", NULL);
	if (!table.fp) {
	    return nil;
	}
	else {
	    long int descfieldnum = table_pos("DATABASE_DESC", table);
	    if (descfieldnum < 0) {
		cerr << "VpfDatabase::description: ";
		cerr << "Invalid DHT (" <<  path()
		     << ") - missing DATABASE_DESC field\n";
		vpf_close_table(&table);
	    }
	    else {
		long int n;
		row = read_next_row(table);
		((VpfDatabase*)this)->_dbdesc =
		    (char *) get_table_element(descfieldnum,row,table,NULL,&n);
		free_row(row,table);
		vpf_close_table(&table);
	    }      
	}
    }
    return _dbdesc;
}

void VpfDatabase::print() const {
    printf("VPF Database %s\n", path());
    printf("Producer: %s\n", _dbprod);
    printf("Libraries: %d\n", _nlibs);
    for (int i = 0; i < _nlibs; i++) {
	printf("Library: %s\n", library(i)->name());
	library(i)->print();
    }
}

char** VpfDatabase::library_names (int& nlibs) const {
    char **libnames;
    vpf_table_type table;
    row_type row;
    char latpath[255];
    strcpy(latpath, path());
    strcat(latpath, "/lat.");
    table = vpf_open_table(latpath, disk, "rb", NULL);
    if (!table.fp) {
	nlibs = 0;
	libnames = NULL;
    }
    else {
	long int libnamefieldnum = table_pos("LIBRARY_NAME", table);
	if (libnamefieldnum < 0) {
	    cerr << "VpfDatabase::library_names: ";
	    cerr << "Invalid LAT (" << path() << ") - missing LIBRARY_NAME field\n";
	    vpf_close_table(&table);
	    nlibs = 0;
	    libnames = NULL;
	}
	else {
	    long int n;
	    libnames = new char*[table.nrows];
	    nlibs = 0;
	    for (int i = 0; i < table.nrows; i++) {
		row = read_next_row(table);
		char* libname;
		libname = (char *)get_table_element(libnamefieldnum,row,table,NULL,&n);
		rightjust (libname);
		strlower(libname);
		free_row(row,table);
		char libpath[255];
		strcpy(libpath, path());
		strcat(libpath, "/");
		struct stat buf;
		strcat(libpath, libname);
		if (stat(libpath,&buf)==0)
		    libnames[nlibs++] = libname;
	    }
	    vpf_close_table(&table);
	}
    }
    return libnames;
}

MapLibrary* VpfDatabase::make_library(const char* name) {
    return new VpfLibrary(this, name);
}

