/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <TopoFace/fgeomobjs.h>
#include <Vpf/vpfcov.h>
#include <Vpf/vpfdb.h>
#include <Vpf/vpflib.h>
#include <Vpf/vpffclass.h>
#include <Vpf/vpffeature.h>

extern "C" {
#include <VpfUtil/vpftidx.h>
}

#include <fstream.h>
#include <string.h>
#include <sys/stat.h>

/*****************************************************************************/

VpfLibrary::VpfLibrary (VpfDatabase* db, const char* name) 
: MapLibrary(db, name) {
    _desc = NULL;
    database(db);
}

VpfLibrary::~VpfLibrary () {
    delete _desc;
}

void VpfLibrary::database(VpfDatabase* db) { 
    _database = (MapDatabase*)db;
    if (_database) {
        init_coverages();
        description();
    }
}

const char* VpfLibrary::path() const {
    if (_database && !_path) {
        struct stat buf;
	const char* libpath = MapLibrary::path();
        if (stat(libpath, &buf) == 0) {
	    char tpath[256];
	    strcpy(tpath, libpath);
	    strcat(tpath, "/tileref");
	    if (stat(tpath, &buf) == 0)
	        ((VpfLibrary*)this)->_tiled = true;
        }
    }
    return _path;
}

VpfCoverage* VpfLibrary::coverage(int covnum) const {
    if (covnum >= 0)
        return (VpfCoverage*)MapLibrary::coverage(covnum);
    else
       return nil;
}

VpfCoverage* VpfLibrary::coverage(const char* name) const {
    return (VpfCoverage*)MapLibrary::coverage(name);
}

const char* VpfLibrary::description() const {
    if (!_desc) {
	vpf_table_type table;
	row_type row;
	char lhtpath[255];
	strcpy(lhtpath, path());
	strcat(lhtpath, "/lht.");
	table = vpf_open_table(lhtpath,disk,"rb",NULL);
	if (!table.fp) 
	    return NULL;
	long int n;
	long int descfieldnum = table_pos("DESCRIPTION",table);
	if (descfieldnum < 0) {
	    cerr << "VpfLibrary::description: Invalid LHT - missing DESCRIPTION field\n";
	    vpf_close_table(&table);
	}
	row = read_next_row(table);
	((VpfLibrary*)this)->_desc = (char *)get_table_element(descfieldnum,row,table,NULL,&n);
	free_row(row,table);
	vpf_close_table(&table);
    }
    return _desc;
}


VpfExtent* VpfLibrary::extent () const {
    /* Get library extent from Library Attribute Table (LAT) */
    char path[256];
    char* buf;
    long int n;
    float xmin,ymin,xmax,ymax;
    int LIBRARY_NAME_, XMIN_, YMIN_, XMAX_, YMAX_, found, i;
    vpf_table_type table;
    row_type row;
    extent_type libextent;
    strcpy( path, database()->path() );
    strcat( path, "/lat." );
    table = vpf_open_table( path, disk, "rb", NULL );
    LIBRARY_NAME_ = table_pos( "LIBRARY_NAME", table );
    found = 0;
    for (i=1;i<=table.nrows;i++) {
	row = read_row( i, table );
	buf = (char *)get_table_element( LIBRARY_NAME_, row, table, NULL, &n );
	rightjust(buf);
	if (stricmp( buf, _name ) == 0) {
	    XMIN_ = table_pos( "XMIN", table );
	    YMIN_ = table_pos( "YMIN", table );
	    XMAX_ = table_pos( "XMAX", table );
	    YMAX_ = table_pos( "YMAX", table );
	    get_table_element( XMIN_, row, table, &xmin, &n );
	    get_table_element( YMIN_, row, table, &ymin, &n );
	    get_table_element( XMAX_, row, table, &xmax, &n );
	    get_table_element( YMAX_, row, table, &ymax, &n );
	    libextent.x1 = xmin;
	    libextent.y1 = ymin;
	    libextent.x2 = xmax;
	    libextent.y2 = ymax;
	    found = 1;
	    i = table.nrows+2;
	}
	free(buf);
	free_row( row, table );
    }
    vpf_close_table( &table );
    return new VpfExtent(libextent);
}

char** VpfLibrary::coverage_names (int& ncovs) const {
    char **covnames;
    vpf_table_type table;
    row_type row;
    char catpath[255];
    strcpy(catpath, path());
    strcat(catpath, "/cat.");
    table = vpf_open_table(catpath, disk, "rb", NULL);
    if (!table.fp) {
	ncovs = 0;
	covnames = NULL;
    }
    else {
	long int covfieldnum = table_pos("COVERAGE_NAME", table);
	if (covfieldnum < 0) {
	    cerr << "VpfLibrary::coverage_names: ";
	    cerr << "Invalid CAT (" << path() << ") - missing COVERAGE_NAME field\n";
	    vpf_close_table(&table);
	    ncovs = 0;
	    covnames = NULL;
	}
	else {
	    long int n;
	    covnames = new char*[table.nrows+2];
	    ncovs = 0;
	    for (int i = 0; i <= table.nrows+1; i++) {
		char* cov;
		if (i == 0) {
		    cov = strdup("tileref");
		}
		else if (i == 1) {
		    cov = strdup("libref");
		}
		else {
		    row = read_next_row(table);
		    cov = (char *)get_table_element(covfieldnum,row,table,NULL,&n);
		    rightjust(cov);
		    strlower(cov);
		}
		char covpath[256];
		strcpy(covpath, path());
		strcat(covpath, "/");
		strcat(covpath, cov);
		struct stat statbuf;
		if (stat(covpath, &statbuf) == 0)
		    covnames[ncovs++] = cov;
		if (i > 1)
		    free_row(row,table);
	    }
	    vpf_close_table(&table);
	}
    }
    return covnames;
}

MapCoverage* VpfLibrary::make_coverage(const char* name) { 
    return new VpfCoverage(this,name);
}
					   

void VpfLibrary::print() const {
    printf(" Library Name: %s\n", _name);
    printf(" Library Path: %s\n", path());
    if (_desc)
	printf(" Description: %s\n", _desc);
    printf(" Coverages: %d\n", _ncovs);
    for (int i = 0; i < _ncovs; i++) {
	printf(" Coverage: %s\n", coverage(i)->name());
	coverage(i)->print();
    }
}

boolean VpfLibrary::gazette() {
    char gfpath[strlen(path())+21];
    char gnpath[strlen(path())+14];
    struct stat buf;

    strcpy(gfpath, path());
    strcat(gfpath, "/gazette/gazette.pft");
    strcpy(gnpath, path());
    strcat(gnpath, "/gazette/end.");
    return (stat(gfpath, &buf)==0 && stat(gnpath, &buf)==0);
}

char* mystrupr(char* str) {
  char* s = str;
  while (*s)
    *s++ = toupper(*s);
  return str;
}

int VpfLibrary::gazette_search(const char* placename, char**& matches, char**& types,
			       float*& posx, float*& posy) {
    int nmatches;
    char gfpath[strlen(path())+21];
    char gnpath[strlen(path())+14];
    struct stat statbuf;
    char* pname = strdup(placename);
    mystrupr(pname);

    strcpy(gfpath, path());
    strcat(gfpath, "/gazette/gazette.pft");
    strcpy(gnpath, path());
    strcat(gnpath, "/gazette/end.");
    if (stat(gfpath, &statbuf)==0 && stat(gnpath, &statbuf)==0) {
	vpf_table_type pft, nodetable;
	point_rec_type point_rec;
	row_type row;
	long int rownum, count, first,last,i;
	int NAME_, TYPE_;
	set_type gaz_set, sel_set;
	int gaz_index_flag = 1;
	char gipath[strlen(path())+21];
	strcpy(gipath, path());
	strcat(gipath, "/gazette/gazette.pti");

	pft = vpf_open_table(gfpath,disk,"rb",NULL);
	gaz_set = set_init(pft.nrows+1);
	NAME_ = table_pos("PLACE_NAME",pft);
	TYPE_ = table_pos("TYPE",pft);
	if (NAME_ < 0) {
	    cerr << "No PLACE_NAME field in gazette.pft\n";
	    vpf_close_table(&pft);
	    return 0;
	}
	if (TYPE_ < 0) {
	    cerr << "No TYPE field in gazette.pft\n";
	    vpf_close_table(&pft);
	    return 0;
	}
	/* Read the gazetteer index, if present, to narrow the search set */
	if (stat(gipath,&statbuf)==0) {
	    sel_set = read_thematic_index(gipath,pname);
	} else {
	    sel_set = set_init(pft.nrows);
	    set_on(sel_set);
	    gaz_index_flag = 0;
	}
	/* Search for place names beginning with 'placename' */
	first = set_min(sel_set);
	if (!gaz_index_flag)
	    last = sel_set.size - 1;
	else
	    last = set_max(sel_set);
	for (rownum=first;rownum<=last;rownum++) {
	    char placestr[80];
	    char* buf;
	    if (!set_member(rownum,sel_set)) continue;
	    row = get_row( rownum+1, pft );
	    buf = (char *)get_table_element(NAME_,row,pft,NULL,&count);
	    strcpy(placestr,buf);
	    free(buf);
	    free_row(row,pft);
	    rightjust(placestr);
	    leftjust(placestr);
	    if (strncmp(mystrupr(placestr), mystrupr(pname), strlen(placename)) == 0)
		set_insert(rownum, gaz_set);
	}
	set_nuke(&sel_set);
	if (set_empty(gaz_set)) {
	    vpf_close_table(&pft);
	    return 0;
	}
	else {
	    char         *buf, *buf1, *buf2;
	    nodetable = vpf_open_table(gnpath,disk,"rb",NULL);
	    first = set_min(gaz_set);
	    last = set_max(gaz_set);
	    nmatches = 0;
	    for (i=first;i<=last;i++) {
		if (!set_member(i,gaz_set)) continue;
		nmatches++;
	    }
	    matches = new char*[nmatches];
	    types = new char*[nmatches];
	    posx = new float[nmatches];
	    posy = new float[nmatches];
	    nmatches = 0;
	    for (i=first;i<=last;i++) {
	       if (!set_member(i,gaz_set)) continue;

	       point_rec = read_point(i+1, nodetable);
	       posx[nmatches] = point_rec.x;
	       posy[nmatches] = point_rec.y;

	       row = get_row( i+1, pft );
	       buf = (char *)get_table_element(TYPE_,row,pft,NULL,&count);
	       rightjust(buf);
	       leftjust(buf);
	       types[nmatches] = strdup(buf);
	       free(buf);
	       buf = (char *)get_table_element(NAME_,row,pft,NULL,&count);
	       rightjust(buf);
	       leftjust(buf);
	       matches[nmatches++] = strdup(buf);
	       free(buf);
	       free_row(row,pft);
	    }
	    vpf_close_table(&pft);
	    vpf_close_table(&nodetable);
	    return nmatches;
	}

    }
    else
	return 0;
}

inline void fixpath(char* path) {
    int i = 0;
    while (path[i]) {
	if (path[i]=='\\')
	    path[i]='/';
	else {
	    int c = path[i];
	    c = tolower(path[i]);
	    path[i] = c;
	}
	i++;
    }
}

inline char *rjust( char *str )
{
   register int  len,i;

   len = strlen(str);
   i = len - 1;
   while ((i>0) && ((str[i]==0) || (str[i]==' '))) i--;
   if (i < (len-1)) str[i+1] = '\0';
   for (i=0;i<strlen(str);i++) if (str[i]=='\n') str[i] = '\0';
   return str;
}

void VpfLibrary::gaz_tileindex() {
  char gfpath[strlen(path())+21];
  char gnpath[strlen(path())+14];
  struct stat statbuf;

  strcpy(gfpath, path());
  strcat(gfpath, "/gazette/gazette.pft");
  strcpy(gnpath, path());
  strcat(gnpath, "/gazette/end.");
  VpfCoverage* cov = coverage("tileref");
  VpfFeatureClass* fc = cov ? cov->feature_class("tileref") : nil;

  if (stat(gfpath, &statbuf)==0 && stat(gnpath, &statbuf)==0 && cov && fc) {
    vpf_table_type pft, nodetable;
    point_rec_type point_rec;
    row_type row, noderow;
    long int rownum, count, first,last,i;
    int NAME_, TYPE_, COORD_;
    char placestr[80];
    char ntypestr[80];
    char* buf;

    pft = vpf_open_table(gfpath,disk,"rb",NULL);
    NAME_ = table_pos("PLACE_NAME",pft);
    TYPE_ = table_pos("TYPE",pft);
    if (NAME_ < 0) {
      cerr << "No PLACE_NAME field in gazette.pft\n";
      vpf_close_table(&pft);
      return;
    }
    if (TYPE_ < 0) {
      cerr << "No TYPE field in gazette.pft\n";
      vpf_close_table(&pft);
      return;
    }

    nodetable = vpf_open_table(gnpath,disk,"rb",NULL);
    COORD_ = table_pos("COORDINATE", nodetable);
    if (COORD_ < 0) {
      cerr << "No COORDINATE field in node table 'end.'\n";
      vpf_close_table(&nodetable);
      return;
    }

    char outname[80];
    sprintf(outname, "gaz_tile_%s", name());
    ofstream outfile(outname);
    for (i = 1; i <= pft.nrows; i++) {
      row = read_row(i, pft);
      buf = (char *)get_table_element(NAME_,row,pft,NULL,&count);
      rjust(buf);
      if (strlen(buf) > 1) {
	strcpy(placestr,buf);
	free(buf);
	buf = (char *)get_table_element(TYPE_,row,pft,NULL,&count);
	rjust(buf);
	strcpy(ntypestr,buf);
	free(buf);
	outfile << placestr << "," << ntypestr;

	point_rec = read_point(i, nodetable);
	FPointObj npoint(point_rec.x, point_rec.y);
	outfile.form(",%f,%f", point_rec.x, point_rec.y);

	for (int j = 1; j < fc->nfeats(); j++) { // avoid first dummy tile
	  VpfFeature* tile = fc->feature(j);
	  VpfExtent* extent = tile->extent();
	  FBoxObj extbox(extent->x1, extent->y1, extent->x2, extent->y2);
	  if (extbox.Contains(npoint)) {
	    char* tname = (char*)tile->named_row_element("TILE_NAME",
							 (void*)NULL,
							 &count);
	    fixpath(tname);
	    outfile << "," << tname;
	    free(tname);
	  }
	}
	outfile << "\n";
      }
      else {
	free(buf);
      }
    }
    vpf_close_table(&pft);
    vpf_close_table(&nodetable);
    outfile.close();
  }
}
